@@ -13,24 +13,29 @@ The following interfaces are currently available:
 # License: Simplified BSD, 2013
 # See LICENSE.txt for more information
 
-from __future__ import print_function
+from __future__ import print_function, unicode_literals
 import warnings
 import abc
 import os
 import atexit
+import logging
+import csv
+from time import time, sleep
 
-from ._py3k_compat import urlopen, quote, HTTPError, iteritems, string_type
+from ._py3k_compat import (urlopen, quote, HTTPError,
+                           iteritems, string_type, reduce)
 
 import re
-import csv
 import numpy as np
 from .scidbarray import SciDBArray, SciDBDataShape, ArrayAlias, SDB_IND_TYPE
-from .errors import SHIM_ERROR_DICT
-from .utils import broadcastable
+from .errors import SHIM_ERROR_DICT, SciDBQueryError, SciDBConnectionError, SciDBInvalidSession
+from .utils import broadcastable, _is_query, iter_record, _new_attribute_label, as_list
+from .schema_utils import disambiguate
 
 __all__ = ['SciDBInterface', 'SciDBShimInterface', 'connect']
 
 SCIDB_RAND_MAX = 2147483647  # 2 ** 31 - 1
+UNESCAPED_QUOTE = re.compile(r"(?<!\\)'")
 
 
 def _df(arr, ind):
@@ -45,29 +50,37 @@ def _af(arr, ind):
     return "{{a.a{i}f}}".format(i=ind).format(a=arr)
 
 
-def _new_attribute_label(suggestion='val', *arrays):
-    """Return a new attribute label
-
-    The label will not clash with any attribute labels in the given arrays
+def _to_bytes(arr):
     """
-    label_list = sum([[dim[0] for dim in arr.sdbtype.full_rep]
-                      for arr in arrays], [])
-    if suggestion not in label_list:
-        return suggestion
-    else:
-        # find all labels of the form val_0, val_1, val_2 ... etc.
-        # where `val` is replaced by suggestion
-        R = re.compile(r'^{0}_(\d+)$'.format(suggestion))
-        nums = sum([map(int, R.findall(label)) for label in label_list], [])
+    Convert a numpy array to a bytestring in SciDB's binary format
+    """
+    # very inefficient like this.
 
-        nums.append(-1)  # in case it's empty
-        return '{0}_{1}'.format(suggestion, max(nums) + 1)
+    # easy case: no strings, SciDB format matches numpy format
+    if not any(np.issubdtype(t, np.character) for l, t in arr.dtype.descr):
+        return arr.tostring(order='C')
+
+    # some attributes are strings
+    result = []
+    for item in arr.ravel():
+        for datum in iter_record(item):
+            dtype = datum.dtype
+            if np.issubdtype(dtype, np.character):
+                datum = datum.astype('U').tostring().decode('utf32').encode('utf8')
+                sz = len(datum) + 1
+                prefix = np.int32(sz).newbyteorder('<').tostring()
+                result.append(prefix + datum + b'\x00')
+            else:
+                result.append(np.array(datum, dtype=dtype).tostring())
+    result = b''.join(result)
+    return result
 
 
 class SciDBInterface(object):
 
     def __init__(self):
         self._created = []
+        self._persistent = set()
         atexit.register(self.reap)
 
     """SciDBInterface Abstract Base Class.
@@ -81,6 +94,7 @@ class SciDBInterface(object):
 
     - ``_execute_query``
     - ``_upload_bytes``
+    - ``_release_session``
     """
     __metaclass__ = abc.ABCMeta
 
@@ -128,6 +142,7 @@ class SciDBInterface(object):
             either a string or a byte string is returned, depending on the
             value of ``fmt``.
         """
+        logging.getLogger(__name__).debug(query)
         if not hasattr(self, '_query_log'):
             self._query_log = []
         self._query_log.append(query)
@@ -142,8 +157,16 @@ class SciDBInterface(object):
             The raw byte data to upload to a file on the SciDB server.
         Returns
         -------
+        (filename, session_id)
+
         filename : string
             The name of the resulting file on the SciDB server.
+        session : object
+            The session ID associated with the file
+
+        It is the responsibility of the caller of _upload_bytes
+        to call _free_session with the returned session object
+        once finished with the uploaded file.
         """
         pass
 
@@ -152,18 +175,26 @@ class SciDBInterface(object):
         """Get a unique query ID from the database"""
         pass
 
+    @abc.abstractmethod
+    def _release_session(self, session):
+        """
+        Close a session with the database.
+        """
+        pass
+
     def reap(self):
         """
         Reap all arrays created via new_array
         """
         for array in list(self._created):
-            if (array.datashape is not None) and (not array.persistent):
-                try:
-                    array.reap()
-                except:
-                    pass
-                finally:
-                    self._created.remove(array)
+            if array in self._persistent:
+                continue
+            try:
+                self.query("remove({0})", array)
+            except SciDBQueryError:  # array does not exist
+                pass
+
+        self._created = []
 
     def _db_array_name(self):
         """Return a unique array name for a new array on the database"""
@@ -178,19 +209,32 @@ class SciDBInterface(object):
             # on subsequent calls, increment the array count
             self.array_count += 1
 
-        return "{0}{1}_{2:05}".format(arr_key, self.uid, self.array_count)
+        result = "{0}{1}_{2:05}".format(arr_key, self.uid, self.array_count)
+        self._created.append(result)
+        return result
 
     def _scan_array(self, name, **kwargs):
         """Return the contents of the given array"""
         if 'response' not in kwargs:
             kwargs['response'] = True
+        if _is_query(name):
+            return self._execute_query(name, **kwargs)
         return self._execute_query("scan({0})".format(name), **kwargs)
 
     def _show_array(self, name, **kwargs):
         """Show the schema of the given array"""
         if 'response' not in kwargs:
             kwargs['response'] = True
-        return self._execute_query("show({0})".format(name), **kwargs)
+        name = UNESCAPED_QUOTE.sub(r"\'", name)
+        if _is_query(name):
+            # need to add a fake store command to trigger
+            # att/dim disambiguation
+            tmp = self._db_array_name()
+            query = "show('store({0}, {1})', 'afl')".format(name, tmp)
+        else:
+            query = "show({0})".format(name)
+
+        return self._execute_query(query, **kwargs)
 
     def _array_dimensions(self, name, **kwargs):
         """Show the dimensions of the given array"""
@@ -266,7 +310,6 @@ class SciDBInterface(object):
         else:
             datashape = None
         result = SciDBArray(datashape, self, name, persistent=persistent)
-        self._created.append(result)
         return result
 
     def _format_query_string(self, query, *args, **kwargs):
@@ -622,15 +665,18 @@ class SciDBInterface(object):
         arr : SciDBArray
             A SciDBArray containint an [n x n] identity matrix
         """
-        arr = self.new_array((n, n), dtype, **kwargs)
-        d0, d1 = arr.dim_names
-        diag = "{0}={1}".format(d0, d1)
+        if len(as_list(dtype)) > 1:
+            raise NotImplementedError("Identity matrices must have 1 attribute")
+        dtype = as_list(dtype)[0]
+
+        query = self.afl.build('<x:%s>[i0=0:%i,1000,0,i1=0:%i,1000,0]' % (dtype, n - 1, n - 1),
+                               'iif(i0=i1,1,0)')
+
         if sparse:
-            query = self.afl.build_sparse(arr, 1, diag)
-        else:
-            query = self.afl.build(arr, 'iif(%s,1,0)' % diag)
-        query.eval(out=arr)
-        return arr
+            # redimension converts NULL to empty
+            query = query.apply('i', 'iif(x=1, i0, NULL)', 'j', 'iif(x=1, i1, NULL)')
+            query = query.redimension('<x:%s>[i=0:%i,1000,0,j=0:%i,1000,0]' % (dtype, n - 1, n - 1))
+        return query.eval()
 
     def dot(self, A, B):
         """Compute the matrix product of A and B
@@ -724,7 +770,7 @@ class SciDBInterface(object):
         ret = []
         for output in ['U', 'S', 'VT']:
             if out_dict[output]:
-                ret.append(self.afl.gesvd(A, "'%s'" % output).eval())
+                ret.append(self.afl.gesvd(A, "'%s'" % output))
         return tuple(ret)
 
     def from_array(self, A, instance_id=0, **kwargs):
@@ -748,10 +794,12 @@ class SciDBInterface(object):
         q = self.afl.quote
         A = np.asarray(A)
         instance_id = int(instance_id)
-        filename = q(self._upload_bytes(A.tostring(order='C')))
+        filename, session_id = self._upload_bytes(_to_bytes(A))
+        filename = q(filename)
         arr = self.new_array(A.shape, A.dtype, **kwargs)
         self.afl.load(arr, filename, instance_id,
                       q(arr.sdbtype.bytes_fmt)).eval(store=False)
+        self._release_session(session_id)
         return arr
 
     def from_dataframe(self, A, instance_id=0, **kwargs):
@@ -823,12 +871,13 @@ class SciDBInterface(object):
         # first flatten the array & indices
         # We'll treat the general case where A.data can be a record array,
         # though this will be very uncommon.
-        flat_dtype = [(d1, SDB_IND_TYPE),
-                      (d2, SDB_IND_TYPE)]
+        flat_dtype = [(str(d1), SDB_IND_TYPE),
+                      (str(d2), SDB_IND_TYPE)]
         if A.dtype.names is None:
-            flat_dtype += [('f0', A.dtype)]
+            flat_dtype += [(str('f0'), A.dtype)]
         else:
             flat_dtype += A.descr
+
         M = np.empty(len(A.data), dtype=flat_dtype)
         M[d1] = A.row
         M[d2] = A.col
@@ -861,12 +910,10 @@ class SciDBInterface(object):
     def _apply_func(self, A, func):
         # TODO: new value name could conflict.  How to generate a unique one?
         # TODO: add optional ``out`` argument as in numpy
-        f = self.afl
         att = A.att(0)
         newatt = "{0}_{1}".format(func, att)
         expr = "{0}({1})".format(func, att)
-        q = f.papply(A, newatt, expr)
-        return q.eval()
+        return self.afl.papply(A, newatt, expr)
 
     def sin(self, A):
         """Element-wise trigonometric sine"""
@@ -904,6 +951,22 @@ class SciDBInterface(object):
         """Element-wise base-10 logarithm"""
         return self._apply_func(A, 'log10')
 
+    def sqrt(self, A):
+        """Element-wise square root"""
+        return self._apply_func(A, 'sqrt')
+
+    def ceil(self, A):
+        """Element-wise ceiling function"""
+        return self._apply_func(A, 'ceil')
+
+    def floor(self, A):
+        """Element-wise floor function"""
+        return self._apply_func(A, 'floor')
+
+    def isnan(self, A):
+        """Element-wise nan test function"""
+        return self._apply_func(A, 'is_nan')
+
     def min(self, A, index=None, scidb_syntax=False):
         """
         Array or axis minimum.
@@ -985,14 +1048,14 @@ class SciDBInterface(object):
     def merge(self, A, B):
         """Merge two arrays"""
         # TODO: pre-check for non-matching arrays?
-        return self.afl.merge(A, B).eval()
+        return self.afl.merge(A, B)
 
     def join(self, *args):
         """
         Perform a series of array joins on the arguments
         and return the result.
         """
-        return reduce(self.afl.join, args).eval()
+        return reduce(self.afl.join, args)
 
     def cross_join(self, A, B, *dims):
         """Perform a cross-join on arrays A and B.
@@ -1007,7 +1070,7 @@ class SciDBInterface(object):
         dims = [_df(arr, index)
                 for dim in dims
                 for arr, index in zip([A, B], dim)]
-        return self.afl.cross_join(A, B, *dims).eval()
+        return self.afl.cross_join(A, B, *dims)
 
     def _join_operation(self, left, right, op):
         """Perform a join operation across arrays or values.
@@ -1015,8 +1078,10 @@ class SciDBInterface(object):
         See e.g. SciDBArray.__add__ for an example usage.
         """
         f = self.afl
+        left, right = disambiguate(left, right)
 
         if isinstance(left, SciDBArray):
+            left = left.eval()
             left_name = left.name
             left_fmt = _af(left, 0)
             left_is_sdb = True
@@ -1026,6 +1091,7 @@ class SciDBInterface(object):
             left_is_sdb = False
 
         if isinstance(right, SciDBArray):
+            right = right.eval()
             right_name = right.name
             right_fmt = _af(right, 0)
             right_is_sdb = True
@@ -1055,10 +1121,9 @@ class SciDBInterface(object):
                 attr = _new_attribute_label('x', left, right)
                 if left_name == right_name:
                     # same array: we can do this without a join
-                    return f.papply(left, attr, op).eval()
+                    return f.papply(left, attr, op)
                 else:
-                    q = f.papply(f.join(left, right), attr, op)
-                    return q.eval()
+                    return f.papply(f.join(left, right), attr, op)
 
             # array shapes are broadcastable: use a cross_join
             elif broadcastable(left.shape, right.shape):
@@ -1087,8 +1152,8 @@ class SciDBInterface(object):
                 if left_slices:
                     aL = ArrayAlias(left, "alias_left")
                     dims = [item for sl in left_slices
-                            for item in [_df(left, sl), 0]]
-                    left_query = f.as_(f.slice(left, *dims), aL).query
+                            for item in [left.dim_names[sl], 0]]
+                    left_query = f.as_(f.slice(left, *dims), aL)
                 else:
                     left_query = left
                     aL = left
@@ -1097,21 +1162,22 @@ class SciDBInterface(object):
                 if right_slices:
                     aR = ArrayAlias(right, "alias_right")
                     dims = [item for sl in right_slices
-                            for item in [_df(right, sl), 0]]
-                    right_query = f.as_(f.slice(right, *dims), aR).query
+                            for item in [right.dim_names[sl], 0]]
+                    right_query = f.as_(f.slice(right, *dims), aR)
                 else:
                     right_query = right
                     aR = right
 
                 # build the cross_join query
-                dims = [_df(arr, att)
-                        for atts in join_indices
-                        for arr, att in zip([aL, aR], atts)]
+                dims = [_df(arr, ind)
+                        for inds in join_indices
+                        for arr, ind in zip([aL, aR], inds)]
+
                 attr = _new_attribute_label('x', left, right)
                 query = f.papply(f.cross_join(
                                  left_query, right_query, *dims),
                                  attr, op)
-                result = query.eval()
+                result = query
 
                 # determine the dimension permutation
                 # Here's the problem: cross_join puts all the left array dims
@@ -1165,7 +1231,7 @@ class SciDBInterface(object):
             except:
                 raise ValueError("rhs must be a scalar or SciDBArray")
             attr = _new_attribute_label('x', left)
-            return f.papply(left, attr, op).eval()
+            return f.papply(left, attr, op)
 
         # only right entry is a SciDBArray
         elif right_is_sdb:
@@ -1174,7 +1240,7 @@ class SciDBInterface(object):
             except:
                 raise ValueError("lhs must be a scalar or SciDBArray")
             attr = _new_attribute_label('x', right)
-            return f.papply(right, attr, op).eval()
+            return f.papply(right, attr, op)
 
         # reorder the dimensions if needed (for cross_join)
         if permutation is not None:
@@ -1233,7 +1299,10 @@ class SciDBShimInterface(SciDBInterface):
 
     def _upload_bytes(self, data):
         session_id = self._shim_new_session()
-        return self._shim_upload_file(session_id, data)
+        return self._shim_upload_file(session_id, data), session_id
+
+    def _release_session(self, session):
+        self._shim_release_session(session)
 
     def _shim_url(self, keyword, **kwargs):
         url = self.hostname + '/' + keyword
@@ -1243,6 +1312,8 @@ class SciDBShimInterface(SciDBInterface):
         return url
 
     def _shim_urlopen(self, url):
+        # shim can freeze up if hit with many connections.
+        # problem often goes away by waiting and retrying
         try:
             return urlopen(url)
         except HTTPError as e:
@@ -1256,25 +1327,32 @@ class SciDBShimInterface(SciDBInterface):
         session_id = int(result.read())
         return session_id
 
-    def _shim_release_session(self, session_id):
+    def _shim_release_session(self, session_id, ignore_invalid=False):
         url = self._shim_url('release_session', id=session_id)
-        self._shim_urlopen(url)
+        if ignore_invalid:
+            try:
+                self._shim_urlopen(url)
+            except SciDBInvalidSession:
+                pass
+        else:
+            self._shim_urlopen(url)
 
     def _shim_execute_query(self, session_id, query, save=None, release=False):
         url = self._shim_url('execute_query',
                              id=session_id,
-                             query=quote(query),
+                             query=quote(query.encode('utf-8')),
                              release=int(bool(release)))
         if save is not None:
             url += "&save={0}".format(quote(save))
 
-        # Don't know the accepted way in Python to do something like this.
-        # For now, just set the 'debug' attribute on this SciDB array object.
-        if hasattr(self, 'debug'):
-            print(query)
+        try:
+            result = self._shim_urlopen(url)
+            query_id = result.read()
+        except KeyboardInterrupt:
+            self._shim_cancel(session_id)
+            self._shim_release_session(session_id, ignore_invalid=True)
+            raise KeyboardInterrupt("Query cancelled")
 
-        result = self._shim_urlopen(url)
-        query_id = result.read()
         return query_id.decode('UTF-8')
 
     def _shim_cancel(self, session_id):
@@ -1283,17 +1361,29 @@ class SciDBShimInterface(SciDBInterface):
 
     def _shim_read_lines(self, session_id, n):
         url = self._shim_url('read_lines', id=session_id, n=n)
+        t0 = time()
         result = self._shim_urlopen(url)
         text_result = result.read()
+        dt = time() - t0
+        pl = len(text_result) / 1048576
+        logging.getLogger(__name__).debug("Transfer time: %0.1f sec", dt)
+        logging.getLogger(__name__).debug("Payload:       %0.2f MB", pl)
+
         # the following check is for Py3K compatibility
         if not isinstance(text_result, string_type):
             text_result = text_result.decode('UTF-8')
         return text_result
 
     def _shim_read_bytes(self, session_id, n):
-        url = self._shim_url('read_lines', id=session_id, n=n)
+        url = self._shim_url('read_bytes', id=session_id, n=n)
+        t0 = time()
         result = self._shim_urlopen(url)
         bytes_result = result.read()
+        dt = time() - t0
+        pl = len(bytes_result) / 1048576
+        logging.getLogger(__name__).debug("Transfer time: %0.1f sec", dt)
+        logging.getLogger(__name__).debug("Payload:       %0.2f MB", pl)
+
         return bytes_result
 
     def _shim_upload_file(self, session_id, data):
