@@ -3,23 +3,29 @@
 # License: Simplified BSD, 2014
 # See LICENSE.txt for more information
 
-from __future__ import print_function
+from __future__ import print_function, division, unicode_literals
+import warnings
+import re
+from itertools import chain
+from functools import partial
 
 import numpy as np
-import re
 
-import warnings
 from copy import copy
-from .errors import SciDBError, SciDBForbidden
+from .errors import SciDBError, SciDBForbidden, SciDBQueryError
 
 # Numpy 1.7 meshgrid backport
-from .utils import meshgrid
-from ._py3k_compat import genfromstr, iteritems
+from . import parse
+from .utils import meshgrid, slice_syntax, _is_query, _new_attribute_label
+from ._py3k_compat import genfromstr, iteritems, csv_reader, string_type
+from .schema_utils import change_axis_schema
 
 __all__ = ["sdbtype", "SciDBArray", "SciDBDataShape"]
 
 
 # Create mappings between scidb and numpy string representations
+# XXX these are partially deprecated, now that nulls/missing values
+# are handled
 _np_typename = lambda s: np.dtype(s).descr[0][1]
 SDB_NP_TYPE_MAP = {'bool': _np_typename('bool'),
                    'float': _np_typename('float32'),
@@ -32,14 +38,58 @@ SDB_NP_TYPE_MAP = {'bool': _np_typename('bool'),
                    'uint16': _np_typename('uint16'),
                    'uint32': _np_typename('uint32'),
                    'uint64': _np_typename('uint64'),
-                   'char': _np_typename('c')}
+                   'char': _np_typename('c'),
+                   'datetime': '<M8[s]',
+                   'datetimetz': '<M8[s]',
+                   'string': '|U100'}
 
 NP_SDB_TYPE_MAP = dict((val, key)
                        for key, val in iteritems(SDB_NP_TYPE_MAP))
 
+INTEGER_TYPES = ('int8', 'int16', 'int32', 'int64',
+                 'uint8', 'uint16', 'uint32', 'uint64')
+
+
+def _sdb_type(np_type):
+    if np.issubdtype(np_type, np.character):
+        return 'string'
+
+    if np_type in NP_SDB_TYPE_MAP:
+        return NP_SDB_TYPE_MAP[np_type]
+
+    raise TypeError("Numpy dtype has no SciDB equivalent: %s" % np_type)
+
 SDB_IND_TYPE = 'int64'
 
 
+def _parse_csv_builtin(txt, dtype):
+    """
+    Convert a SciDB-output csv document into a NumPy array
+
+    Uses the builtin CSV module to handle string dtypes
+    """
+    # TODO: optional pandas version, for speed
+
+    if not any(np.issubdtype(t, np.character) for f, t in dtype.descr):
+        return genfromstr(txt, skip_header=1, delimiter=',', dtype=dtype)
+
+    r = csv_reader(txt, delimiter=',', quotechar="'",
+                   escapechar='\\', skiplines=1)
+    r = list(map(tuple, r))
+
+    # resize string dtypes to accommodate longest string
+    new_dtype = [f if not np.issubdtype(f[1], np.character)
+                 else (f[0], 'U%i' % max(len(row[i]) for row in r))
+                 for i, f in enumerate(dtype.descr)]
+    if len(new_dtype) == 1:
+        new_dtype = new_dtype[0][1]
+    dtype = np.dtype(new_dtype)
+
+    return np.array(r, dtype=dtype)
+
+_parse_csv = _parse_csv_builtin
+
+
 class sdbtype(object):
 
     """SciDB data type class.
@@ -92,7 +142,8 @@ class sdbtype(object):
     @property
     def bytes_fmt(self):
         """The format used for transfering raw bytes to and from scidb"""
-        return '({0})'.format(','.join(rep[1] for rep in self.full_rep))
+        return "(%s)" % ','.join(f if not n else "%s NULL" % f
+                                 for (_, f, n) in self.full_rep)
 
     @classmethod
     def _regularize(cls, schema):
@@ -126,11 +177,11 @@ class sdbtype(object):
         # assume that now we just have the dtypes themselves
         # TODO: support default values?
         sdbL = schema.split(',')
-        sdbL = [list(map(str.strip, s.split(':'))) for s in sdbL]
+        sdbL = [list(map(lambda x: x.strip(), s.split(':'))) for s in sdbL]
 
-        names = [s[0] for s in sdbL]
+        names = [str(s[0]) for s in sdbL]
         dtypes = [s[1].split()[0] for s in sdbL]
-        nullable = ['null' in str.lower(''.join(s[1].split()[1:]))
+        nullable = ['null' in (''.join(s[1].split()[1:])).lower()
                     for s in sdbL]
         return list(zip(names, dtypes, nullable))
 
@@ -150,7 +201,6 @@ class sdbtype(object):
             The corresponding numpy dtype descriptor
         """
         sdbL = cls._schema_to_list(schema)
-
         if len(sdbL) == 1:
             return np.dtype(SDB_NP_TYPE_MAP[sdbL[0][1]])
         else:
@@ -175,7 +225,7 @@ class sdbtype(object):
         # Hack: if we re-encode this as a dtype, then de-encode again, numpy
         #       will add default names where they are missing
         dtype = np.dtype(dtype).descr
-        pairs = ["{0}:{1}".format(d[0], NP_SDB_TYPE_MAP[d[1]]) for d in dtype]
+        pairs = ["{0}:{1}".format(d[0], _sdb_type(d[1])) for d in dtype]
         return '<{0}>'.format(','.join(pairs))
 
 
@@ -184,12 +234,26 @@ class SciDBDataShape(object):
     """Object to store SciDBArray data type and shape"""
 
     def __init__(self, shape, typecode, dim_names=None,
-                 chunk_size=1000, chunk_overlap=0):
+                 chunk_size=1000, chunk_overlap=0, dim_low=None,
+                 dim_high=None):
+
         # Process array shape
-        try:
-            self.shape = tuple(shape)
-        except:
-            self.shape = (shape,)
+        if shape is not None:
+            try:
+                shape = tuple(shape)
+            except:
+                shape = (shape,)
+            dim_low = [0] * len(shape)
+            dim_high = [s - 1 for s in shape]
+
+        if dim_low is None or dim_high is None:
+            raise ValueError("Must specify dim_low and dim_high, or shape")
+
+        dim_low = tuple(map(lambda x: None if x == '*' else int(x), dim_low))
+        dim_high = tuple(map(lambda x: None if x == '*' else int(x), dim_high))
+
+        self.dim_low = dim_low
+        self.dim_high = dim_high
 
         # process array typecode: either a scidb schema or a numpy dtype
         self.sdbtype = sdbtype(typecode)
@@ -210,27 +274,39 @@ class SciDBDataShape(object):
             dim_names = ['i{0}'.format(i)
                          for i in range(start, start + len(self.shape))]
 
-        if len(dim_names) != len(self.shape):
+        if len(dim_names) != len(self.dim_high):
             raise ValueError("length of dim_names should match "
                              "number of dimensions")
         self.dim_names = dim_names
 
+        ndim = len(dim_names)
+
         # process chunk sizes.  Either an integer, or a list of integers
         if not hasattr(chunk_size, '__len__'):
-            chunk_size = [chunk_size for s in self.shape]
-        if len(chunk_size) != len(self.shape):
+            chunk_size = [chunk_size] * ndim
+        if len(chunk_size) != ndim:
             raise ValueError("length of chunk_size should match "
                              "number of dimensions")
         self.chunk_size = chunk_size
 
         # process chunk overlaps.  Either an integer, or a list of integers
         if not hasattr(chunk_overlap, '__len__'):
-            chunk_overlap = [chunk_overlap for s in self.shape]
-        if len(chunk_overlap) != len(self.shape):
+            chunk_overlap = [chunk_overlap] * ndim
+        if len(chunk_overlap) != ndim:
             raise ValueError("length of chunk_overlap should match "
                              "number of dimensions")
         self.chunk_overlap = chunk_overlap
 
+        if len(self.dim_names) != len(set(self.dim_names)):
+            warnings.warn("Duplicate dimension names: %s" % self.dim_names)
+
+        if len(self.sdbtype.names) != len(set(self.sdbtype.names)):
+            warnings.warn("Duplicate attribute names: %s" % self.sdbtype.names)
+
+    @property
+    def ndim(self):
+        return len(self.dim_names)
+
     @classmethod
     def from_schema(cls, schema):
         """Create a DataShape object from a SciDB Schema.
@@ -265,24 +341,60 @@ class SciDBDataShape(object):
             raise ValueError("no match for schema: {0}".format(schema))
 
         # split dshapes.  TODO: correctly handle '*' dimensions
-        #                       handle non-integer dimensions?
-        R = re.compile(r'(\S*?)=(\S*?):(\S*?),(\S*?),(\S*?),')
+        R = re.compile(r'(\S*?)=(\S*?):([\S\*]*?),(\S*?),(\S*?),')
         dshapes = R.findall(dshapes + ',')  # note added trailing comma
 
-        return cls(shape=[int(d[2]) - int(d[1]) + 1 for d in dshapes],
+        return cls(None,
                    typecode=schema,
                    dim_names=[d[0] for d in dshapes],
                    chunk_size=[int(d[3]) for d in dshapes],
-                   chunk_overlap=[int(d[4]) for d in dshapes])
+                   chunk_overlap=[int(d[4]) for d in dshapes],
+                   dim_low=[d[1] for d in dshapes],
+                   dim_high=[d[2] for d in dshapes])
+
+    @classmethod
+    def from_query(cls, interface, query):
+        """
+        Build a datashape from an AFL query string
+
+        Parameters
+        ----------
+        interface : SciDBInterface
+        query : str
+
+        Returns
+        -------
+        A SciDBDataShape instance, inferred from the database
+        """
+        schema = interface._show_array(query)
+        schema = schema.split("'")[1]
+        return cls.from_schema(schema)
+
+    @property
+    def shape(self):
+        if None in self.dim_low or None in self.dim_high:
+            return None
+
+        return tuple(h - l + 1 for l, h in zip(self.dim_low, self.dim_high))
 
     @property
     def schema(self):
-        shape_arg = ','.join(['{0}=0:{1},{2},{3}'.format(d, s - 1, cs, co)
-                              for (d, s, cs, co) in zip(self.dim_names,
-                                                        self.shape,
+        return '{0} {1}'.format(self.sdbtype, self.dim_schema)
+
+    @property
+    def dim_schema(self):
+        """
+        The dimension part of the schema
+        """
+        result = ','.join(['{0}={1}:{2},{3},{4}'.format(d, l,
+                                                        h if h is not None else '*',
+                                                        cs, co)
+                           for (d, l, h, cs, co) in zip(self.dim_names,
+                                                        self.dim_low,
+                                                        self.dim_high,
                                                         self.chunk_size,
                                                         self.chunk_overlap)])
-        return '{0} [{1}]'.format(self.sdbtype, shape_arg)
+        return '[%s]' % result
 
     @property
     def ind_attr_dtype(self):
@@ -294,7 +406,7 @@ class SciDBDataShape(object):
         dct = dict(f[:2] for f in self.sdbtype.full_rep)
         types = [SDB_NP_TYPE_MAP[dct.get(key, SDB_IND_TYPE)]
                  for key in keys]
-        return np.dtype(list(zip(keys, types)))
+        return np.dtype(list(zip(map(str, keys), types)))
 
 
 class ArrayAlias(object):
@@ -364,6 +476,39 @@ class SciDBArray(object):
         self.name = name
         self.persistent = persistent
 
+    @property
+    def persistent(self):
+        """ Controls whether the array is deleted when
+        the database is reaped
+        """
+        return self.name in self.interface._persistent
+
+    @persistent.setter
+    def persistent(self, value):
+        if value:
+            self.interface._persistent.add(self.name)
+        elif self.name in self.interface._persistent:
+            self.interface._persistent.remove(self.name)
+
+    @classmethod
+    def from_query(cls, interface, query):
+        """
+        Build a lazily-evaulated SciDB array from a query string
+
+        Parameters
+        ----------
+        interface : SciDBInterface
+            The database connection to use
+        query : str
+            The query string to wrap
+
+        Returns
+        --------
+        array : SciDBArray
+        """
+        result = cls(None, interface, query)
+        return result
+
     @property
     def afl(self):
         """
@@ -371,6 +516,47 @@ class SciDBArray(object):
         """
         return self.interface.afl
 
+    @property
+    def schema(self):
+        """Return the array schema"""
+        return self.datashape.schema
+
+    def head(self, n=5):
+        """
+        Extract and download the first few elements in the array
+
+        Parameters
+        ----------
+        n : int (optional, default=5)
+           The number of elements to retrieve
+
+        Returns
+        -------
+        head : SciDBArray
+            The first N elements in the array, downloaded
+            as a Pandas dataframe (if pandas is installed) or a Numpy array
+
+        """
+        if self.shape is not None:
+            n = min(n, self.size)
+
+        args = zip([0] * (self.ndim - 1), [1] * (self.ndim - 1))
+        result = self.subarray(0, n - 1, *args)
+        try:
+            return result.todataframe()
+        except ImportError:
+            return result.toarray()
+
+    def tail(self, n=5):
+        hi = self.size - 1
+        lo = max(hi - n + 1, 0)
+        args = zip([0] * (self.ndim - 1), [1] * (self.ndim - 1))
+        result = self.subarray(lo, hi, *args)
+        try:
+            return result.todataframe()
+        except ImportError:
+            return result.toarray()
+
     def rename(self, new_name, persistent=False):
         """Rename the array in the database, optionally making the new
         array persistent.
@@ -462,6 +648,8 @@ class SciDBArray(object):
             try:
                 schema = self.interface._show_array(self.name, fmt='csv')
                 self._datashape = SciDBDataShape.from_schema(schema)
+            except SciDBQueryError as exc:
+                raise SciDBQueryError("Invalid query:\n\n%s\n\n%s" % (self.name, exc))
             except SciDBError:
                 self._datashape = None
         return self._datashape
@@ -482,9 +670,13 @@ class SciDBArray(object):
     def dim_names(self):
         return self.datashape.dim_names
 
+    @property
+    def att_names(self):
+        return self.sdbtype.names
+
     @property
     def ndim(self):
-        return len(self.datashape.shape)
+        return self.datashape.ndim
 
     @property
     def size(self):
@@ -498,6 +690,12 @@ class SciDBArray(object):
     def dtype(self):
         return self.datashape.dtype
 
+    @property
+    def query(self):
+        if _is_query(self.name):
+            return self.name
+        return None
+
     def reap(self, ignore=False):
         """
         Delete this object from the database if it isn't persistent.
@@ -512,6 +710,9 @@ class SciDBArray(object):
         ------
         SciDBForbidden if ``persistent=True`` and ``ignore=False`
         """
+        if _is_query(self.name):
+            return
+
         if self.persistent:
             if not ignore:
                 raise SciDBForbidden("Cannot reap: persistent=True")
@@ -625,6 +826,10 @@ class SciDBArray(object):
         if output not in ['auto', 'dense', 'sparse']:
             raise ValueError("unrecognized output: '{0}'".format(output))
 
+        # workaround for strings
+        if any(s[1] == 'string' for s in self.sdbtype.full_rep):
+            transfer_bytes = False
+
         dtype = self.datashape.dtype
         sdbtype = self.sdbtype
         shape = self.datashape.shape
@@ -655,9 +860,7 @@ class SciDBArray(object):
                 raise ValueError("Fatal: unexpected array labels.")
 
             # convert the ASCII representation into a numpy record array
-            arr = np.atleast_1d(genfromstr(str_rep, delimiter=',',
-                                           skip_header=1,
-                                           dtype=full_dtype))
+            arr = np.atleast_1d(_parse_csv(str_rep, full_dtype))
 
             if transfer_bytes:
                 # replace parsed ASCII columns with more accurate bytes
@@ -669,24 +872,29 @@ class SciDBArray(object):
                         arr[name] = bytes_arr[name]
 
             if output == 'dense':
-                if self.ndim != 2:
-                    raise NotImplementedError("sparse to dense for ndim != 2")
-                from scipy import sparse
-                data = arr[full_dtype.names[2]]
-                ij = (arr[full_dtype.names[0]], arr[full_dtype.names[1]])
-                arr_coo = sparse.coo_matrix((data, ij), shape=self.shape)
-                arr = arr_coo.toarray()
+                result = np.zeros(self.shape, self.dtype)
+                coords = tuple([arr[d] for d in self.dim_names])
+
+                if len(self.att_names) == 1:
+                    result[coords] = arr[self.att_names[0]]
+                    return result
+
+                for att in self.att_names:
+                    result[att][coords] = arr[att]
+                return result
 
         else:
             if transfer_bytes:
                 # reshape bytes_array to the correct shape
-                arr = bytes_arr.reshape(shape)
+                try:
+                    arr = bytes_arr.reshape(shape)
+                except ValueError:
+                    arr = bytes_arr
             else:
                 # transfer via ASCII.  Here we don't need the indices, so we
                 # use 'csv' output.
                 str_rep = self.interface._scan_array(self.name, n=0, fmt='csv')
-                arr = genfromstr(str_rep, delimiter=',', skip_header=1,
-                                 dtype=dtype).reshape(shape)
+                arr = _parse_csv(str_rep, dtype).reshape(shape)
 
             if output == 'sparse':
                 index_arrays = list(map(np.ravel,
@@ -703,27 +911,70 @@ class SciDBArray(object):
 
         return arr
 
-    def toarray(self, transfer_bytes=True):
+    def toarray(self, transfer_bytes=False):
         """Transfer data from database and store in a numpy array.
 
         Parameters
         ----------
-        transfer_bytes : boolean
-            if True (default), then transfer data as bytes rather than as
-            ASCII.
+        transfer_bytes : DEPRECATED
+           Unused
 
         Returns
         -------
         arr : np.ndarray
             The dense array containing the data.
+
+        Notes
+        -----
+        If the array is backed by a query, the query is evaluated and stored
+        in the database
         """
-        return self._download_data(transfer_bytes=transfer_bytes,
-                                   output='dense')
+        if transfer_bytes:
+            warnings.warn(DeprecationWarning("transfer_bytes is deprecated, "
+                          "and will be removed in a future version"))
+
+        self.eval()  # evaluate if needed, for speed
+        return parse.toarray(self)
+
+    def eval(self, out=None, store=True, **kwargs):
+        """
+        If the array is backed by an unevaluated query,
+        evaluate the query and store the result in the database
+
+        This changes array.name from a query string to a
+        stored array name. Calling eval() on an array
+        that is already backed by a stored array does nothing.
+
+        Parameters
+        ----------
+        out : SciDBArray (optional)
+           An optional pre-existing array to store the evaluation into.
+        """
+        if not _is_query(self.name):
+            return self
+
+        if not store:
+            return self.interface._execute_query(self.name, **kwargs)
+
+        if out is not None:
+            self.persistent = out.persistent
+            name = out.name
+            result = out
+        else:
+            name = self.interface._db_array_name()
+            result = self
+
+        query = 'store({q}, {name})'.format(q=self.name, name=name)
+        self.interface._execute_query(query, **kwargs)
+
+        self.name = name
+        return result
 
     def todataframe(self, transfer_bytes=True):
         """Transfer array from database and store in a local Pandas dataframe
 
-        This is valid only for a one-dimensional array.
+        For multidimensional arrays, the dimension values are added
+        as additional columns in the dataframe.
 
         Parameters
         ----------
@@ -736,10 +987,12 @@ class SciDBArray(object):
         arr : pd.DataFrame
             The dataframe object containing the data in the array.
         """
-        if self.ndim != 1:
-            raise ValueError("Dataframe export is valid only for 1D arrays")
         from pandas import DataFrame
-        return DataFrame(self.toarray())
+        if self.ndim == 1:
+            return DataFrame(self.toarray())
+
+        idx = _new_attribute_label('row', self)
+        return self.afl.unpack(self, idx).todataframe()
 
     def tosparse(self, sparse_fmt='recarray', transfer_bytes=True):
         """Transfer array from database and store in a local sparse array.
@@ -812,11 +1065,27 @@ class SciDBArray(object):
         # TODO: allow newaxis to be passed
 
         # slice can be either a tuple/iterable or a single integer/slice
+
+        if isinstance(indices, SciDBArray):
+            return self._boolean_filter(indices)
+
+        # passing a boolean mask
+        if isinstance(indices, np.ndarray) and indices.dtype == np.bool:
+            indices = self.interface.from_array(indices)
+            return self._boolean_filter(indices)
+
+        if isinstance(indices, string_type):
+            indices = (indices,)
+
         try:
             indices = tuple(indices)
         except TypeError:
             indices = (indices,)
 
+        # accessing attributes by name
+        if all(isinstance(i, string_type) for i in indices):
+            return self.afl.project(self, *indices)
+
         if len(indices) > self.ndim:
             raise ValueError("too many indices")
 
@@ -826,6 +1095,11 @@ class SciDBArray(object):
         # if num indices is less than num dimensions, right-fill them
         indices = list(indices) + [slice(None)] * (self.ndim - len(indices))
 
+        # special case, indexing with N 1D SciDB arrays, return
+        # row/column/etc subset where these arrays are nonempty
+        if all(isinstance(i, SciDBArray) for i in indices):
+            return _subarray(self, *indices)
+
         # special case: accessing a single element (no slices)
         if all(not isinstance(i, slice) for i in indices):
             limits = list(map(int, indices + indices))
@@ -838,7 +1112,7 @@ class SciDBArray(object):
                   if not isinstance(s, slice)
                   for item in (d, s)]
         if slices:
-            arr1 = self.afl.slice(self, *slices).eval()
+            arr1 = self.afl.slice(self, *slices)
         else:
             arr1 = self
 
@@ -850,19 +1124,41 @@ class SciDBArray(object):
         # if a subarray is required, then call the subarray() command
         if any(i[0] != 0 or i[1] != s for (i, s) in zip(indices, shape)):
             limits = [i[0] for i in indices] + [i[1] - 1 for i in indices]
-            arr2 = self.afl.subarray(arr1, *limits).eval()
+            arr2 = self.afl.subarray(arr1, *limits)
         else:
             arr2 = arr1
 
         # if thinning is required, then call the thin() command
         if any(i[2] != 1 for i in indices):
             steps = sum([[0, i[2]] for i in indices], [])
-            arr3 = self.afl.thin(arr2, *steps).eval()
+            arr3 = self.afl.thin(arr2, *steps)
         else:
             arr3 = arr2
 
         return arr3
 
+    def __setitem__(self, key, value):
+        if isinstance(key, string_type):
+            key = [key]
+            value = [value]
+        if len(key) != len(value):
+            raise ValueError("Number of expressions does not match number "
+                             "of new attributes")
+        args = chain(*zip(key, value))
+        result = self.afl.apply(self, *args)
+        self.name = result.name
+        self._datashape = None  # refresh schema
+
+    @slice_syntax
+    def sdbslice(self, slices):
+        try:
+            slices = tuple(slices)
+        except TypeError:
+            slices = (slices,)
+
+        args = [s.start for s in slices] + [s.stop - 1 for s in slices]
+        return self.afl.subarray(self, *args)
+
     # join operations: note that these ignore all but the first attribute
     # of each array.
     def __add__(self, other):
@@ -912,6 +1208,121 @@ class SciDBArray(object):
     def __abs__(self):
         return self.interface._apply_func(self, 'abs')
 
+    def __getattr__(self, attr):
+        """
+        Fallback getattribute
+
+        If attr is the name of an AFL operator,
+        we apply self to that operator and return the partial result
+
+        Examples:
+        ---------
+        x.filter('f0 > 3') # -> afl.filter(x, 'f0>3')
+        """
+        if hasattr(self.afl, attr):
+            return partial(getattr(self.afl, attr), self)
+        raise AttributeError(attr)
+
+    def _boolean_compare(self, operator, other):
+        """
+        Build a new column based on an inequality test
+        """
+        if isinstance(other, np.ndarray):
+            other = self.interface.from_array(other)
+
+        if isinstance(other, SciDBArray):
+            return self._boolean_compare_array(operator, other)
+
+        if len(self.att_names) > 1:
+            raise TypeError("Inequality comparison not supported for "
+                            "multi-attribute arrays")
+
+        f = self.afl
+
+        # wrap strings for convenience
+        if isinstance(other, string_type) and not other.startswith("'"):
+            other = "'%s'" % other
+
+        expr = '{att} {op} {other}'.format(att=self.att(0),
+                                           other=other,
+                                           op=operator)
+        att = _new_attribute_label('condition', self)
+        return f.papply(self, att, expr)
+
+    def _boolean_compare_array(self, operator, other):
+        """
+        Test operator(self, other), where other is a SciDBArray
+        """
+        from .schema_utils import disambiguate
+
+        if len(self.att_names) > 1 or len(other.att_names) > 1:
+            raise TypeError("Inequality comparison not supported for "
+                            "multi-attribute arrays")
+
+        if self.shape != other.shape:
+            raise ValueError("Array shapes do not match")
+
+        self, other = disambiguate(self, other)
+
+        f = self.afl
+        joined = f.join(self, other)
+        att = _new_attribute_label('condition', joined)
+        expr = "{self} {op} {other}".format(self=self.att_names[0],
+                                            op=operator,
+                                            other=other.att_names[0])
+        return f.papply(joined, att, expr)
+
+    def _boolean_filter(self, mask):
+        """
+        Extract flattened array of elements in self where mask is true.
+
+        Equivalent to numpy mask filtering: x[mask]
+        """
+        from .schema_utils import disambiguate
+
+        if mask.shape and self.shape and mask.shape != self.shape:
+            raise ValueError("Shape of mask does not match array: %s vs %s" %
+                             (mask.shape, self.shape))
+
+        if len(mask.att_names) != 1:
+            raise TypeError("Boolean mask must have a single attribute")
+
+        f = self.afl
+        self, mask = disambiguate(self, mask)
+        joined = f.join(self, mask)
+        expr = '%s=TRUE' % mask.att_names[0]
+        idx = _new_attribute_label('__idx', self, mask)
+        return f.project(f.unpack(f.filter(joined, expr), idx),
+                         *self.att_names)
+
+    def __lt__(self, other):
+        return self._boolean_compare('<', other)
+
+    def __le__(self, other):
+        return self._boolean_compare('<=', other)
+
+    def __eq__(self, other):
+        return self._boolean_compare('=', other)
+
+    def __ne__(self, other):
+        return self._boolean_compare('!=', other)
+
+    def __ge__(self, other):
+        return self._boolean_compare('>=', other)
+
+    def __gt__(self, other):
+        return self._boolean_compare('>', other)
+
+    def __invert__(self):
+        if len(self.att_names) != 1:
+            raise TypeError("Can only invert single-attribute arrays")
+        if self.sdbtype.full_rep[0][1] != 'bool':
+            raise TypeError("Can only invert boolean arrays")
+
+        newatt = _new_attribute_label('condition', self)
+        att = self.att_names[0]
+        return self.afl.papply(self, newatt, "not(%s)" % att)
+
     def transpose(self, *axes):
         """Permute the dimensions of an array.
 
@@ -944,7 +1355,7 @@ class SciDBArray(object):
                     pass
 
         if not axes:
-            arr = self.afl.transpose(self).eval()
+            arr = self.afl.transpose(self)
         else:
             # first validate the axes
             axes = [(a + self.ndim if a < 0 else a) for a in axes]
@@ -985,6 +1396,19 @@ class SciDBArray(object):
         arr : SciDBArray
             new array of the specified shape
         """
+        if isinstance(shape, int):
+            shape = (shape,)
+
+        # handle -1s (see :meth:`numpy.reshape`)
+        if any(s == -1 for s in shape):
+            n = np.prod([s for s in shape if s != -1])
+            ntot = np.prod(self.shape)
+            if (ntot // n) * n != ntot:
+                raise ValueError("total size of new array "
+                                 "must remain unchanged")
+            shape = list(shape)
+            shape[shape.index(-1)] = ntot // n
+
         if np.prod(shape) != np.prod(self.shape):
             raise ValueError("new shape is incompatible")
         arr = self.interface.new_array(shape=shape,
@@ -1003,10 +1427,34 @@ class SciDBArray(object):
         -------
         arr : SciDBArray
             new non-nullable array
+
+        Notes
+        -----
+        This is currently limited to single-attribute arrays.
+        Use the raw AFL substutute operator for multi-attribute arrays
         """
         b = self.afl.build('%s[i=0:0,1,0]' % self.sdbtype, value)
         q = self.afl.substitute(self, b)
-        return q.eval()
+        return q
+
+    def groupby(self, by):
+        """
+        Build a groupby object from this array
+
+        Parameters
+        ----------
+        by : string or list of strings
+            Names of attributes and dimensions to group by
+
+        Returns
+        -------
+        groups : :class:`scidbpy.aggregation.GroupBy` instance
+            An object that can be used, e.g., to perform
+            aggregations over each group. See :class:`scidbpy.aggregation.GroupBy`
+            documentation for more information.
+        """
+        from .aggregation import GroupBy
+        return GroupBy(self, by)
 
     def _aggregate_operation(self, agg, index=None, scidb_syntax=True):
         """Perform an aggregation query
@@ -1068,8 +1516,7 @@ class SciDBArray(object):
                 idx_args = [self.dim_names[i] for i in ind]
 
         agg = "{agg}({att})".format(agg=agg, att=self.att(0))
-        q = self.afl.aggregate(self, agg, *idx_args)
-        return q.eval()
+        return self.afl.aggregate(self, agg, *idx_args)
 
     def min(self, index=None, scidb_syntax=False):
         """
@@ -1317,5 +1764,175 @@ class SciDBArray(object):
 
         agg = "{agg}({att})".format(agg=aggregate, att=self.att(0))
         args = list(map(str, sizes)) + [agg]
-        q = self.afl.regrid(self, *args)
-        return q.eval()
+        return self.afl.regrid(self, *args)
+
+    def cumulate(self, expression, dimension=0):
+        """
+        Compute running operations along data (e.g., cumulative sums)
+
+        Parameters
+        ----------
+        expression: str
+            A valid SciDB expression
+        dimension : int or str (optional, default=0)
+           Which dimension to accumulate over
+
+        Returns
+        -------
+        arr : SciDBArray
+            A new array of the same shape.
+
+        Examples
+        --------
+        >>> x = sdb.arange(12).reshape((3, 4))
+        >>> x.cumulate('sum(f0)').toarray()
+        array([[ 0,  1,  2,  3],
+              [ 4,  6,  8, 10],
+              [12, 15, 18, 21]])
+
+        See Also
+        --------
+        cumsum(), cumprod()
+        """
+        if isinstance(dimension, int):
+            dimension = self.dim_names[dimension]
+
+        return self.afl.cumulate(self, expression, dimension)
+
+    def cumsum(self, axis=None):
+        """
+        Return the cumulative sum over the array.
+
+        Parameters
+        ----------
+        axis : int, optional
+           The axis to sum over. The default sums over the
+           flattened array
+
+        Returns
+        -------
+        sums : SciDBArray
+            A new array, with the same shape (but flattened if axis=None)
+
+        See Also
+        --------
+        cumprod(), cumulate()
+        """
+        return self._agg_ufunc('sum', axis)
+
+    def cumprod(self, axis=None):
+        """
+        Return the cumulative product over the array.
+
+        Parameters
+        ----------
+        axis : int, optional
+           The axis to multiply over. The default multiplies over the
+           flattened array
+
+        Returns
+        -------
+        prods : SciDBArray
+            A new array, with the same shape (but flattened if axis=None)
+
+        See Also
+        --------
+        cumsum(), cumulate()
+        """
+        return self._agg_ufunc('prod', axis)
+
+    def _agg_ufunc(self, func, axis):
+        if axis is None:
+            self = self.reshape((self.size,))
+            axis = self.dim_names[0]
+        if isinstance(axis, int):
+            axis = self.dim_names[axis]
+
+        sums = ["%s(%s) as %s" % (func, att, att)
+                for att in self.att_names]
+        sums = ", ".join(sums)
+        return self.afl.cumulate(self, sums, axis)
+
+    def compress(self, mask, axis=0):
+        """
+        Extract a subset of entries along a given axis,
+        where an input mask array is non-null
+
+        Parameters
+        ----------
+        array : SciDBArray
+            The array to filter
+        mask : SciDBArray
+            A 1-dimensional SciDBArray, whose non-null values indicate
+            the entries to retain
+        axis : int
+            The axis of array along which to apply the mask. The shape
+            of array along this axis must be the length of mask
+        """
+        f = self.afl
+
+        if mask.sdbtype.full_rep[0][1] == 'bool':
+            mask = f.filter(mask, "%s=TRUE" % mask.att_names[0])
+
+        dim = self.dim_names[axis]
+        chunk = self.datashape.chunk_size[axis]
+        overlap = self.datashape.chunk_overlap[axis]
+        sz = self.shape[axis]
+        ct = int(f.aggregate(mask, 'count(*)').eval()[0])
+        if ct == 0:
+            raise ValueError("Cannot discard all elements of array")
+
+        new_att = _new_attribute_label(dim, mask)
+        idx_att = _new_attribute_label(new_att + '_0', mask)
+
+        # copy the dimension to a new attribute, sort moves nulls to end
+        q = f.papply(mask, new_att, mask.dim_names[0])
+        q = f.sort(q, '%s asc' % new_att)
+
+        # rename attributes, and swap sorted new_att to a dimension
+        # after this step, idx_att contains the final location for
+        # each original location
+        q = f.cast(q, '<%s:int64>[%s=0:*,1000,0]' % (new_att, idx_att))
+        q = f.redimension(q, '<{idx_att}:int64>[{new_att}=0:{stop},{chunk},{overlap}]'
+                          .format(new_att=new_att, stop=sz - 1,
+                                  chunk=chunk, overlap=overlap,
+                                  idx_att=idx_att))
+
+        # tack on new location for each element
+        q = f.cross_join(f.as_(self, 'xj1'),
+                         f.as_(q, 'xj2'),
+                         'xj1.%s' % dim, 'xj2.%s' % new_att)
+
+        # promote idx_att to a dimension, which rearranges + truncates
+        schema = self.sdbtype.schema
+        schema += change_axis_schema(self.datashape,
+                                     axis, name=idx_att, stop=ct - 1).dim_schema
+        q = f.redimension(q, schema)
+        return q
+
+
+def _subarray(array, *masks):
+    """
+    Return a row/column/etc subset of an Nd array, by dropping
+    null locations in N 1D masks
+
+    Parameters
+    ----------
+    array : SciDBArray
+        The array to filter
+
+    masks : tuple of 1D SciDBArrays
+       The axis filters
+
+    Returns
+    -------
+    out : SciDBArray
+       The filtered subarray
+    """
+
+    # XXX perform shape checking here
+
+    result = array
+    for i, mask in enumerate(masks):
+        result = result.compress(result, mask, i).eval()
+    return result
