@@ -1,3 +1,6 @@
+# License: Simplified BSD, 2014
+# See LICENSE.txt for more information
+
 import numpy as np
 from scidbpy.scidbarray import SDB_NP_TYPE_MAP, sdbtype
 
@@ -10,8 +13,3 @@ def test_sdbtype_dtype_mapping():
                  for j in range(i, min(len(type_list), i + 3))]
         dtype_start = np.dtype(dtype)
         assert(dtype_start == sdbtype(dtype_start).dtype)
-
-
-if __name__ == '__main__':
-    import nose
-    nose.runmodule()
